#include "i2cmaster.h"

/**
 * Initialize I2C bus interface.
 */
void I2C_init(void)
{
	/* initialize TWI clock: 100 kHz clock, TWPS = 0 => prescaler = 1 */
	TWSR = 0;

	/* must be > 10 for stable operation */
	TWBR = ((F_CPU / SCL_CLOCK) - 16) / 2;
}

/**  
 * Issues a start condition and sends address and transfer direction.
 *
 * @param address - device address
 *
 * @return 0 - device accessible, 1 - failed to access device
 */
unsigned char I2C_start(unsigned char address)
{
	uint32_t I2C_timer = 0;
	uint8_t twst;

	/* send START condition */
	TWCR = (1 << TWINT) | (1 << TWSTA) | (1 << TWEN);

	/* wait until transmission completed */
	I2C_timer = I2C_TIMER_DELAY;
	while (!(TWCR & (1 << TWINT)) && I2C_timer--) ;

	if (I2C_timer == 0)
		return 1;

	/* check value of TWI status register and mask prescaler bits */
	twst = TW_STATUS & 0xF8;
	if ((twst != TW_START) && (twst != TW_REP_START))
		return 1;

	/* send device address */
	TWDR = address;
	TWCR = (1 << TWINT) | (1 << TWEN);

	/* wail until transmission completed and ACK/NACK has been received */
	I2C_timer = I2C_TIMER_DELAY;
	while (!(TWCR & (1<<TWINT)) && I2C_timer--) ;

	if (I2C_timer == 0)
		return 1;

	/* check value of TWI Status Register. Mask prescaler bits */
	twst = TW_STATUS & 0xF8;
	if ((twst != TW_MT_SLA_ACK) && (twst != TW_MR_SLA_ACK))
		return 1;

	return 0;
}

/**
 * Issues a start condition and sends address and transfer direction.
 * If device is busy, use ack polling to wait until device is ready.
 *
 * @param address - device address
 */
void I2C_start_wait(unsigned char address)
{
	uint32_t I2C_timer = 0;
	uint8_t twst;

	while (1) {
		/* send START condition */
		TWCR = (1 << TWINT) | (1 << TWSTA) | (1 << TWEN);

		/* wait until transmission completed */
		I2C_timer = I2C_TIMER_DELAY;
		while (!(TWCR & (1 << TWINT)) && I2C_timer--) ;

		/* check value of TWI status register and mask prescaler bits */
		twst = TW_STATUS & 0xF8;
		if ((twst != TW_START) && (twst != TW_REP_START))
			continue;

		/* send device address */
		TWDR = address;
		TWCR = (1 << TWINT) | (1 << TWEN);

		/* wail until transmission completed */
		I2C_timer = I2C_TIMER_DELAY;
		while (!(TWCR & (1<<TWINT)) && I2C_timer--) ;

		/* check value of TWI status register and mask prescaler bits */
		twst = TW_STATUS & 0xF8;
		if ((twst == TW_MT_SLA_NACK ) || (twst ==TW_MR_DATA_NACK)) {           
			/*
			 * device busy, send stop condition to
			 * terminate write operation
			 */
			TWCR = (1 << TWINT) | (1 << TWEN) | (1 << TWSTO);

			/*
			 * wait until stop condition is
			 * executed and bus released
			 */
			I2C_timer = I2C_TIMER_DELAY;
			while ((TWCR & (1 << TWSTO)) && I2C_timer--) ;

			continue;
		}

		break;
	}

}

/**
 * Terminates the data transfer and releases the I2C bus.
 */
void I2C_stop(void)
{
	uint32_t I2C_timer = 0;

	/* send stop condition */
	TWCR = (1 << TWINT) | (1 << TWEN) | (1 << TWSTO);

	// wait until stop condition is executed and bus released
	I2C_timer = I2C_TIMER_DELAY;
	while ((TWCR & (1 << TWSTO)) && I2C_timer--) ;
}

/**
 * Send one byte to I2C device.
 *
 * @param data - byte to be transfered
 *
 * @return 0 = write successful, 1 = write failed
 */
unsigned char I2C_wr(unsigned char data)
{   
	uint32_t I2C_timer = 0;
	uint8_t twst;

	/* send data to the previously addressed device */
	TWDR = data;
	TWCR = (1 << TWINT) | (1 << TWEN);

	/* wait until transmission completed */
	I2C_timer = I2C_TIMER_DELAY;
	while (!(TWCR & (1 << TWINT)) && I2C_timer--) ;
	
	if (I2C_timer == 0)
		return 1;

	/* check value of TWI status register and mask prescaler bits */
	twst = TW_STATUS & 0xF8;
	
	if (twst != TW_MT_DATA_ACK)
		return 1;
	
	return 0;
}

/**
 * Read one byte from the I2C device, request more data from device. 
 *
 * @return byte read from I2C device
 */
unsigned char I2C_rd_ack(void)
{
	uint32_t I2C_timer = 0;

	TWCR = (1 << TWINT) | (1 << TWEN) | (1 << TWEA);
	
	I2C_timer = I2C_TIMER_DELAY;
	while (!(TWCR & (1 << TWINT)) && I2C_timer--) ;
	
	if (I2C_timer == 0)
		return 0;

	return TWDR;

}

/**
 * Read one byte from the I2C device, read is followed by a stop condition.
 *
 * @return byte read from I2C device
 */
unsigned char I2C_rd_nak(void)
{
	uint32_t  I2C_timer = 0;

	TWCR = (1 << TWINT) | (1 << TWEN);

	I2C_timer = I2C_TIMER_DELAY;
	while (!(TWCR & (1<<TWINT)) && I2C_timer--) ;

	if (I2C_timer == 0)
		return 0;

	return TWDR;
}
