#ifndef I2C_H
#define I2C_H_		1

#include <inttypes.h>
#include <compat/twi.h>

/**
 * Define CPU frequency in Mhz here if not defined
 * in Makefile or other modules
 */
#ifndef F_CPU
#define F_CPU		16000000UL
#endif

/* I2C clock in Hz */
#define SCL_CLOCK 	400000L

/* I2C timer max delay */
#define I2C_TIMER_DELAY 0xFF

#define WB		0
#define RB		1

/**
 * Initialize I2C bus interface.
 */
extern void I2C_init(void);

/**  
 * Issues a start condition and sends address and transfer direction.
 *
 * @param address - device address
 *
 * @return 0 - device accessible, 1 - failed to access device
 */
extern unsigned char I2C_start(unsigned char address);

/**
 * Issues a start condition and sends address and transfer direction.
 * If device is busy, use ack polling to wait until device is ready.
 *
 * @param address - device address
 */
extern void I2C_start_wait(unsigned char address);

/**
 * Terminates the data transfer and releases the I2C bus.
 */
extern void I2C_stop(void);

/**
 * Send one byte to I2C device.
 *
 * @param data - byte to be transfered
 *
 * @return 0 = write successful, 1 = write failed
 */
extern unsigned char I2C_wr(unsigned char data);

/**
 * Read one byte from the I2C device, request more data from device. 
 *
 * @return byte read from I2C device
 */
extern unsigned char I2C_rd_ack(void);

/**
 * Read one byte from the I2C device, read is followed by a stop condition.
 *
 * @return byte read from I2C device
 */
extern unsigned char I2C_rd_nak(void);

/* useful defines for reading and writing data from / to I2C bus */

/**
 * Read data once from I2C device.
 */
#define I2C_read(ret, adr, reg)					\
	I2C_start((adr << 1) | WB); I2C_wr(reg);		\
	I2C_start((adr << 1) | RB); ret = I2C_rd_nak();		\
	I2C_stop()

/**
 * Read data multiple times from I2C device.
 */
#define I2C_read_multiple(ret, adr, reg, count, ii)			\
	I2C_start((adr << 1) | WB); I2C_wr(reg);		\
	I2C_stop();						\
	I2C_start((adr << 1) | RB);				\
								\
	ii = 0;						\
	while (ii < count - 1)					\
		ret[ii++] = I2C_rd_ack();			\
	ret[ii] = I2C_rd_nak();					\
								\
	I2C_stop()

/**
 * Write data to I2C device.
 */
#define I2C_write(adr, reg, dat)				\
	I2C_start((adr << 1) | WB); I2C_wr(reg);		\
	I2C_wr(dat);						\
	I2C_stop()

#endif	/* I2C_H_ */
