#define __PROG_TYPES_COMPAT__
#define F_CPU 16000000L

#include "lcd.h"

#include <avr/pgmspace.h>
#include <avr/io.h>
#include <avr/interrupt.h>

#include <limits.h>
#include <util/delay.h>

static void SPI_init(void)
{
	SPI_REG |= (1 << LCD_CMD);
	SPI_REG |= (1 << LCD_RST);
	SPI_REG |= (1 << LCD_CS);
	SPI_REG |= (1 << LCD_SCK);
	SPI_REG |= (1 << LCD_MOSI);
	SPI_REG |= (1 << PB4);

	SPI_PORT &= ~(1 << LCD_CS);
	SPI_PORT &= ~(1 << LCD_SCK);
	SPI_PORT &= ~(1 << LCD_MOSI);
	SPI_PORT |= (1 << PB4);

	SPCR |= (1 << MSTR) | (1 << SPE);
}

static void SPI_end(void)
{
	SPCR &= ~(1 << SPE);
}

static inline void SPI_write(uint8_t byte)
{
	SPDR = byte;

	while (!(SPSR & (1 << SPIF)))
		;
}

static void SPI_write_cmd(uint8_t cmd)
{
	SPI_PORT &= ~(1 << LCD_CMD);
	SPI_write(cmd);
}

static void SPI_write_data(uint8_t data)
{
	SPI_PORT |= (1 << LCD_CMD);
	SPI_write(data);
}

// Taken from https://github.com/sumotoy/TFT_ILI9163C/blob/master/TFT_ILI9163C.cpp#L337
PROGMEM static const prog_uchar
TFT141B_CMD_INIT[] = {
	21,
	TFT141_SWRESET, TFT141_DELAY,
	255,
	TFT141_SLPOUT, TFT141_DELAY,
	5,
	TFT141_COLMOD, TFT141_DELAY + 1,
	0x05, 5,
	TFT141_GSET, TFT141_DELAY + 1,
	0x08, 1,
	TFT141_GAMRSEL, TFT141_DELAY + 1,
	0x01, 1,
	TFT141_NORON, 0,
	TFT141_DISSET5, 2,
	0b11111111, 0b00000110,
	TFT141_GMCTRP1, 15,
	0x3F,0x26,0x23,0x30,0x28,0x10,0x55,0xB7,0x40,0x19,0x10,0x1E,0x02,0x01,0x00,
	TFT141_GMCTRN1, 15,
	0x09,0x18,0x2D,0x0D,0x13,0x15,0x40,0x48,0x53,0x0C,0x1D,0x25,0x2E,0x24,0x29,
	TFT141_FRMCTR1, TFT141_DELAY + 2,
	0x08, 0x02, 1,
	TFT141_INVCTR, TFT141_DELAY + 1,
	0x07, 1,
	TFT141_PWCTR1, TFT141_DELAY + 2,
	0x0A, 0x02, 1,
	TFT141_PWCTR2, TFT141_DELAY + 1,
	0x02, 1,
	TFT141_VMCTR1, TFT141_DELAY + 2,
	0x50, 0x5b, 1,
	TFT141_VMCTROFF, TFT141_DELAY + 1,
	0x00, 1,
	TFT141_CASET, 2,
	0x00, TFT141_TFT_WIDTH,
	TFT141_RASET, 2,
	0x00, TFT141_TFT_HEIGHT + 32, // needs 160
	TFT141_VERTSCRL, 3,
	32, TFT141_TFT_HEIGHT, 0,
	TFT141_DISPON, TFT141_DELAY,
	1,
	TFT141_MADCTL, 1,
	// rotation info: https://github.com/sumotoy/TFT_ILI9163C/blob/master/TFT_ILI9163C.cpp#L871-L905
	// 3rd bit is always RGB: 0, GBR - 1
	0b00000000,
	TFT141_RAMWR, TFT141_DELAY,
	1,
};

static void parse_cmd(const uint8_t *addr)
{
	uint8_t num_cmd, num_args;
	uint16_t ms;

	num_cmd = pgm_read_byte(addr++);
	while (num_cmd--) {
		SPI_write_cmd(pgm_read_byte(addr++));
		num_args = pgm_read_byte(addr++);

		ms = num_args & TFT141_DELAY;
		num_args &= ~TFT141_DELAY;

		while (num_args--)
			SPI_write_data(pgm_read_byte(addr++));

		if (ms) {
			ms = pgm_read_byte(addr++);

			if(ms == 255)
				ms = 500;

			_delay_ms(ms);
		}
	}
}

static void set_backlight(void)
{
	SPI_REG |= (1 << LCD_BL);

	TCCR0A |= (1 << COM0A1) | (1 << WGM00);
	TCCR0B |= (1 << CS00);

	OCR0A = TFT141_BRIGHTNESS;
}

static int display_width = 128;
static int display_height = 160;
static uint8_t col, row;

void LCD_init(void)
{
	col = row = 0;

	SPI_init();

	SPI_PORT |= (1 << LCD_RST);
	_delay_ms(100);
	SPI_PORT &= ~(1 << LCD_RST);
	_delay_ms(100);
	SPI_PORT |= (1 << LCD_RST);
	_delay_ms(100);
	
	parse_cmd(TFT141B_CMD_INIT);

	set_backlight();	
}

void LCD_end(void)
{
	SPI_end();
}

void LCD_set_range(uint16_t x1, uint16_t y1, uint16_t x2, uint16_t y2)
{
	SPI_write_cmd(TFT141_CASET);

	SPI_write_data(0x00);
	SPI_write_data(x1 + col);
	SPI_write_data(0x00);
	SPI_write_data(x2 + col);

	SPI_write_cmd(TFT141_RASET);

	SPI_write_data(0x00);
	SPI_write_data(y1 + row);
	SPI_write_data(0x00);
	SPI_write_data(y2 + row);

	SPI_write_cmd(TFT141_RAMWR);
}

void LCD_set_pixel(uint16_t color)
{
	SPI_PORT |= (1 << LCD_CMD);
	
	SPI_write(color >> 8);
        SPI_write(color);
}

void LCD_draw_pixel(int16_t x, int16_t y, uint16_t color)
{
	if ((x < 0) || (x >= display_width) ||
	    (y < 0) || (y >= display_height))
		return;

	LCD_set_range(x, y, x + 1, y + 1);
	LCD_set_pixel(color);
}

void LCD_fill_screen(uint16_t color)
{
	uint8_t x, y;

	LCD_set_range(0, 0, display_width - 1, display_height - 1);

	for (y = display_height; y > 0; y--)
		for (x = display_width; x > 0; x--)
			LCD_set_pixel(color);
}

void LCD_invert_display(uint8_t inv)
{
	SPI_write_cmd(inv ? TFT141_INVON : TFT141_INVOFF);
}

uint16_t LCD_get_color(uint8_t r, uint8_t g, uint8_t b)
{
	return ((r & 0xF8) << 8) | ((g & 0xFC) << 3) | (b >> 3);
}
